@@ -39,7 +39,6 @@
 #include "nv_include.h"
 #include "nv_dma.h"
 
-
 #include "vl_hwmc.h"
 
 #include "hwdefs/nv_m2mf.xml.h"
@@ -196,6 +195,100 @@ static XF86ImageRec NVImages[NUM_IMAGES_ALL] =
 	XVIMAGE_RGB
 };
 
+static void
+nouveau_box_intersect(BoxPtr dest, BoxPtr a, BoxPtr b)
+{
+    dest->x1 = a->x1 > b->x1 ? a->x1 : b->x1;
+    dest->x2 = a->x2 < b->x2 ? a->x2 : b->x2;
+    dest->y1 = a->y1 > b->y1 ? a->y1 : b->y1;
+    dest->y2 = a->y2 < b->y2 ? a->y2 : b->y2;
+
+    if (dest->x1 >= dest->x2 || dest->y1 >= dest->y2)
+	dest->x1 = dest->x2 = dest->y1 = dest->y2 = 0;
+}
+
+static void
+nouveau_crtc_box(xf86CrtcPtr crtc, BoxPtr crtc_box)
+{
+    if (crtc->enabled) {
+	crtc_box->x1 = crtc->x;
+	crtc_box->x2 = crtc->x + xf86ModeWidth(&crtc->mode, crtc->rotation);
+	crtc_box->y1 = crtc->y;
+	crtc_box->y2 = crtc->y + xf86ModeHeight(&crtc->mode, crtc->rotation);
+    } else
+	crtc_box->x1 = crtc_box->x2 = crtc_box->y1 = crtc_box->y2 = 0;
+}
+
+static int
+nouveau_box_area(BoxPtr box)
+{
+    return (int) (box->x2 - box->x1) * (int) (box->y2 - box->y1);
+}
+
+xf86CrtcPtr
+nouveau_pick_best_crtc(ScrnInfoPtr pScrn, Bool consider_disabled,
+                       int x, int y, int w, int h)
+{
+    xf86CrtcConfigPtr   xf86_config = XF86_CRTC_CONFIG_PTR(pScrn);
+    int                 coverage, best_coverage, c;
+    BoxRec              box, crtc_box, cover_box;
+    RROutputPtr         primary_output = NULL;
+    xf86CrtcPtr         best_crtc = NULL, primary_crtc = NULL;
+
+    if (!pScrn->vtSema)
+	return NULL;
+
+    box.x1 = x;
+    box.x2 = x + w;
+    box.y1 = y;
+    box.y2 = y + h;
+    best_coverage = 0;
+
+    /* Prefer the CRTC of the primary output */
+#ifdef HAS_DIXREGISTERPRIVATEKEY
+    if (dixPrivateKeyRegistered(rrPrivKey))
+#endif
+    {
+	primary_output = RRFirstOutput(pScrn->pScreen);
+    }
+    if (primary_output && primary_output->crtc)
+	primary_crtc = primary_output->crtc->devPrivate;
+
+    /* first consider only enabled CRTCs */
+    for (c = 0; c < xf86_config->num_crtc; c++) {
+	xf86CrtcPtr crtc = xf86_config->crtc[c];
+
+	if (!crtc->enabled)
+	    continue;
+
+	nouveau_crtc_box(crtc, &crtc_box);
+	nouveau_box_intersect(&cover_box, &crtc_box, &box);
+	coverage = nouveau_box_area(&cover_box);
+	if (coverage > best_coverage ||
+	    (coverage == best_coverage && crtc == primary_crtc)) {
+	    best_crtc = crtc;
+	    best_coverage = coverage;
+	}
+    }
+    if (best_crtc || !consider_disabled)
+	return best_crtc;
+
+    /* if we found nothing, repeat the search including disabled CRTCs */
+    for (c = 0; c < xf86_config->num_crtc; c++) {
+	xf86CrtcPtr crtc = xf86_config->crtc[c];
+
+	nouveau_crtc_box(crtc, &crtc_box);
+	nouveau_box_intersect(&cover_box, &crtc_box, &box);
+	coverage = nouveau_box_area(&cover_box);
+	if (coverage > best_coverage ||
+	    (coverage == best_coverage && crtc == primary_crtc)) {
+	    best_crtc = crtc;
+	    best_coverage = coverage;
+	}
+    }
+    return best_crtc;
+}
+
 unsigned int
 nv_window_belongs_to_crtc(ScrnInfoPtr pScrn, int x, int y, int w, int h)
 {
@@ -206,7 +299,7 @@ nv_window_belongs_to_crtc(ScrnInfoPtr pScrn, int x, int y, int w, int h)
 	for (i = 0; i < xf86_config->num_crtc; i++) {
 		xf86CrtcPtr crtc = xf86_config->crtc[i];
 
-		if (!crtc->enabled)
+		if (!drmmode_crtc_on(crtc))
 			continue;
 
 		if ((x < (crtc->x + crtc->mode.HDisplay)) &&
@@ -257,10 +350,10 @@ nouveau_xv_bo_realloc(ScrnInfoPtr pScrn, unsigned flags, unsigned size,
 	}
 
 	if (flags & NOUVEAU_BO_VRAM) {
-		if (pNv->Architecture == NV_ARCH_50)
+		if (pNv->Architecture == NV_TESLA)
 			config.nv50.memtype = 0x70;
 		else
-		if (pNv->Architecture >= NV_ARCH_C0)
+		if (pNv->Architecture >= NV_FERMI)
 			config.nvc0.memtype = 0xfe;
 	}
 	flags |= NOUVEAU_BO_MAP;
@@ -729,7 +822,7 @@ NV_calculate_pitches_and_mem_size(NVPtr pNv, int action_flags, int *srcPitch,
 {
 	int tmp;
 
-	if (pNv->Architecture >= NV_ARCH_50) {
+	if (pNv->Architecture >= NV_TESLA) {
 		npixels = (npixels + 7) & ~7;
 		nlines = (nlines + 7) & ~7;
 	}
@@ -975,6 +1068,8 @@ NVPutImage(ScrnInfoPtr pScrn, short src_x, short src_y, short drw_x,
 	if (pPriv->grabbedByV4L)
 		return Success;
 
+	if (width > pPriv->max_image_dim || height > pPriv->max_image_dim)
+		return BadMatch;
 
 	NV_set_action_flags(pScrn, pDraw, pPriv, id, drw_x, drw_y, drw_w,
 			    drw_h, &action_flags);
@@ -1080,7 +1175,7 @@ NVPutImage(ScrnInfoPtr pScrn, short src_x, short src_y, short drw_x,
 			pPriv->TT_mem_chunk[pPriv->currentHostBuffer];
 	}
 	if (!destination_buffer) {
-		if (pNv->Architecture >= NV_ARCH_50) {
+		if (pNv->Architecture >= NV_TESLA) {
 			NOUVEAU_ERR("No scratch buffer for tiled upload\n");
 			return BadAlloc;
 		}
@@ -1283,7 +1378,7 @@ CPU_copy:
 						  src_w, src_h, drw_w, drw_h,
 						  clipBoxes, ppix, pPriv);
 		} else
-		if (pNv->Architecture == NV_ARCH_50) {
+		if (pNv->Architecture == NV_TESLA) {
 			ret = nv50_xv_image_put(pScrn, pPriv->video_mem,
 						offset, uv_offset,
 						id, dstPitch, &dstBox, 0, 0,
@@ -1348,11 +1443,6 @@ NVQueryImageAttributes(ScrnInfoPtr pScrn, int id,
 {
 	int size, tmp;
 
-	if (*w > IMAGE_MAX_W)
-		*w = IMAGE_MAX_W;
-	if (*h > IMAGE_MAX_H)
-		*h = IMAGE_MAX_H;
-
 	*w = (*w + 1) & ~1; // width rounded up to an even number
 	if (offsets)
 		offsets[0] = 0;
@@ -1613,6 +1703,7 @@ NVSetupBlitVideo (ScreenPtr pScreen)
 	pPriv->bicubic			= FALSE;
 	pPriv->doubleBuffer		= FALSE;
 	pPriv->SyncToVBlank		= (pNv->dev->chipset >= 0x11);
+	pPriv->max_image_dim            = 2048;
 
 	pNv->blitAdaptor		= adapt;
 
@@ -1673,6 +1764,7 @@ NVSetupOverlayVideoAdapter(ScreenPtr pScreen)
 	pPriv->blitter			= FALSE;
 	pPriv->texture			= FALSE;
 	pPriv->bicubic			= FALSE;
+	pPriv->max_image_dim            = 2048;
 
 	NVSetPortDefaults (pScrn, pPriv);
 
@@ -1874,6 +1966,7 @@ NV30SetupTexturedVideo (ScreenPtr pScreen, Bool bicubic)
 	pPriv->bicubic			= bicubic;
 	pPriv->doubleBuffer		= FALSE;
 	pPriv->SyncToVBlank		= TRUE;
+	pPriv->max_image_dim            = 4096;
 
 	if (bicubic)
 		pNv->textureAdaptor[1]	= adapt;
@@ -1955,6 +2048,7 @@ NV40SetupTexturedVideo (ScreenPtr pScreen, Bool bicubic)
 	pPriv->bicubic			= bicubic;
 	pPriv->doubleBuffer		= FALSE;
 	pPriv->SyncToVBlank		= TRUE;
+	pPriv->max_image_dim            = 4096;
 
 	if (bicubic)
 		pNv->textureAdaptor[1]	= adapt;
@@ -2048,7 +2142,7 @@ NVSetupTexturedVideo (ScreenPtr pScreen, XF86VideoAdaptorPtr *textureAdaptor)
 		textureAdaptor[0] = NV40SetupTexturedVideo(pScreen, FALSE);
 		textureAdaptor[1] = NV40SetupTexturedVideo(pScreen, TRUE);
 	} else
-	if (pNv->Architecture >= NV_ARCH_50) {
+	if (pNv->Architecture >= NV_TESLA) {
 		textureAdaptor[0] = NV50SetupTexturedVideo(pScreen);
 	}
 }
@@ -2079,10 +2173,10 @@ NVInitVideo(ScreenPtr pScreen)
 	 * might work without accel, we also disable it for now when
 	 * acceleration is disabled:
 	 */
-	if (pScrn->bitsPerPixel != 8 && !pNv->NoAccel) {
+	if (pScrn->bitsPerPixel != 8 && pNv->AccelMethod == EXA) {
 		xvSyncToVBlank = MAKE_ATOM("XV_SYNC_TO_VBLANK");
 
-		if (pNv->Architecture < NV_ARCH_50) {
+		if (pNv->Architecture < NV_TESLA) {
 			overlayAdaptor = NVSetupOverlayVideo(pScreen);
 			blitAdaptor    = NVSetupBlitVideo(pScreen);
 		}
